/**
 * @fileoverview 百度地图的鼠标绘制工具，对外开放。
 * 允许用户在地图上点击完成鼠标绘制的功能。
 * 使用者可以自定义所绘制结果的相关样式，例如线宽、颜色、测线段距离、面积等等。
 * 主入口类是<a href="symbols/DrawingManager"></a>，
 * 基于Baidu Map API 1.0。
 *
 * @author Baidu Map Api Group 
 * @version 1.0
 */


window['BMapLib'] = window['BMapLib'] || {};
window['BMapLib']['EventWrapper'] = window['BMapLib']['EventWrapper'] || {};

(function (root, factory) {
    if (typeof exports === 'object') {
        module.exports = { default: factory(), DrawingManager: factory() };
    } else if (typeof define === 'function' && define.amd) {
        define(factory);
    } else {
        root.BMapLib = root.BMapLib || {};
        root.BMapLib.DrawingManager = root.BMapLib.DrawingManager || factory();
    }
})(this, function () {

    var EventWrapper = window['BMapLib']['EventWrapper'];

    EventWrapper['addDomListener'] = function (instance, eventName, handler) {
        if (instance.addEventListener) {
            instance.addEventListener(eventName, handler, false);
        }
        else if (instance.attachEvent) {
            instance.attachEvent('on' + eventName, handler);
        }
        else {
            instance['on' + eventName] = handler;
        }
        return new MapsEventListener(instance, eventName, handler, MapsEventListener.DOM_EVENT);
    };

    EventWrapper['addDomListenerOnce'] = function (instance, eventName, handler) {
        var eventListener = EventWrapper['addDomListener'](instance, eventName, function () {
            EventWrapper['removeListener'](eventListener);
            return handler.apply(this, arguments);
        });
        return eventListener;
    };

    EventWrapper['addListener'] = function (instance, eventName, handler) {
        instance.addEventListener(eventName, handler);
        return new MapsEventListener(instance, eventName, handler, MapsEventListener.MAP_EVENT);
    };

    EventWrapper['addListenerOnce'] = function (instance, eventName, handler) {
        var eventListener = EventWrapper['addListener'](instance, eventName, function () {
            EventWrapper['removeListener'](eventListener);
            return handler.apply(this, arguments);
        });
        return eventListener;
    };

    EventWrapper['clearInstanceListeners'] = function (instance) {
        var listeners = instance._e_ || {};
        for (var i in listeners) {
            EventWrapper['removeListener'](listeners[i]);
        }
        instance._e_ = {};
    };

    EventWrapper['clearListeners'] = function (instance, eventName) {
        var listeners = instance._e_ || {};
        for (var i in listeners) {
            if (listeners[i]._eventName == eventName) {
                EventWrapper['removeListener'](listeners[i]);
            }
        }
    };

    EventWrapper['removeListener'] = function (listener) {
        var instance = listener._instance;
        var eventName = listener._eventName;
        var handler = listener._handler;
        var listeners = instance._e_ || {};
        for (var i in listeners) {
            if (listeners[i]._guid == listener._guid) {
                if (listener._eventType == MapsEventListener.DOM_EVENT) {
                    if (instance.removeEventListener) {
                        instance.removeEventListener(eventName, handler, false);
                    }
                    else if (instance.detachEvent) {
                        instance.detachEvent('on' + eventName, handler);
                    }
                    else {
                        instance['on' + eventName] = null;
                    }
                }
                else if (listener._eventType == MapsEventListener.MAP_EVENT) {
                    instance.removeEventListener(eventName, handler);
                }
                delete listeners[i];
            }
        }
    };

    EventWrapper['trigger'] = function (instance, eventName) {
        var listeners = instance._e_ || {};
        for (var i in listeners) {
            if (listeners[i]._eventName == eventName) {
                var args = Array.prototype.slice.call(arguments, 2);
                listeners[i]._handler.apply(instance, args);
            }
        }
    };

    function MapsEventListener(instance, eventName, handler, eventType) {

        this._instance = instance;
        this._eventName = eventName;
        this._handler = handler;
        this._eventType = eventType;
        this._guid = MapsEventListener._guid++;
        this._instance._e_ = this._instance._e_ || {};
        this._instance._e_[this._guid] = this;
    }
    MapsEventListener._guid = 1;

    MapsEventListener.DOM_EVENT = 1;
    MapsEventListener.MAP_EVENT = 2;

    var DRAWING_MODE_MARKER = "marker",
        DRAWING_MODE_CIRCLE = "circle";
    DRAWING_MODE_AREA = "area";
    DRAWING_MODE_FITBOUNDS = "fitbounds";

    function extend(subClass, superClass, className) {
        var key, proto,
            selfProps = subClass.prototype,
            clazz = new Function();

        clazz.prototype = superClass.prototype;
        proto = subClass.prototype = new clazz();
        for (key in selfProps) {
            proto[key] = selfProps[key];
        }

        subClass.prototype.constructor = subClass;
        subClass.superClass = superClass.prototype;
        if ("string" == typeof className) {
            proto._className = className;
        }
    };

    var T, baidu = T = baidu || { version: '1.5.0' };
    baidu.guid = '$BAIDU$';
    window[baidu.guid] = window[baidu.guid] || {};

    baidu.lang = baidu.lang || {};
    baidu.lang.isString = function (source) {
        return '[object String]' == Object.prototype.toString.call(source);
    };
    baidu.lang.isFunction = function (source) {
        return '[object Function]' == Object.prototype.toString.call(source);
    };
    baidu.lang.Event = function (type, target) {
        this.type = type;
        this.returnValue = true;
        this.target = target || null;
        this.currentTarget = null;
    };


    baidu.object = baidu.object || {};
    baidu.extend =
        baidu.object.extend = function (target, source) {
            for (var p in source) {
                if (source.hasOwnProperty(p)) {
                    target[p] = source[p];
                }
            }

            return target;
        };
    baidu.event = baidu.event || {};
    baidu.event._listeners = baidu.event._listeners || [];
    baidu.dom = baidu.dom || {};

    baidu.dom._g = function (id) {
        if (baidu.lang.isString(id)) {
            return document.getElementById(id);
        }
        return id;
    };
    baidu._g = baidu.dom._g;
    baidu.event.on = function (element, type, listener) {
        type = type.replace(/^on/i, '');
        element = baidu.dom._g(element);
        var realListener = function (ev) {

            listener.call(element, ev);
        },
            lis = baidu.event._listeners,
            filter = baidu.event._eventFilter,
            afterFilter,
            realType = type;
        type = type.toLowerCase();
        if (filter && filter[type]) {
            afterFilter = filter[type](element, type, realListener);
            realType = afterFilter.type;
            realListener = afterFilter.listener;
        }

        if (element.addEventListener) {
            element.addEventListener(realType, realListener, false);
        } else if (element.attachEvent) {
            element.attachEvent('on' + realType, realListener);
        }
        lis[lis.length] = [element, type, listener, realListener, realType];
        return element;
    };

    baidu.on = baidu.event.on;
    baidu.event.un = function (element, type, listener) {
        element = baidu.dom._g(element);
        type = type.replace(/^on/i, '').toLowerCase();

        var lis = baidu.event._listeners,
            len = lis.length,
            isRemoveAll = !listener,
            item,
            realType, realListener;
        while (len--) {
            item = lis[len];

            if (item[1] === type
                && item[0] === element
                && (isRemoveAll || item[2] === listener)) {
                realType = item[4];
                realListener = item[3];
                if (element.removeEventListener) {
                    element.removeEventListener(realType, realListener, false);
                } else if (element.detachEvent) {
                    element.detachEvent('on' + realType, realListener);
                }
                lis.splice(len, 1);
            }
        }

        return element;
    };
    baidu.un = baidu.event.un;
    baidu.dom.g = function (id) {
        if ('string' == typeof id || id instanceof String) {
            return document.getElementById(id);
        } else if (id && id.nodeName && (id.nodeType == 1 || id.nodeType == 9)) {
            return id;
        }
        return null;
    };
    baidu.g = baidu.G = baidu.dom.g;
    baidu.dom._styleFixer = baidu.dom._styleFixer || {};
    baidu.dom._styleFilter = baidu.dom._styleFilter || [];
    baidu.dom._styleFilter.filter = function (key, value, method) {
        for (var i = 0, filters = baidu.dom._styleFilter, filter; filter = filters[i]; i++) {
            if (filter = filter[method]) {
                value = filter(key, value);
            }
        }
        return value;
    };
    baidu.string = baidu.string || {};

    baidu.string.toCamelCase = function (source) {
        if (source.indexOf('-') < 0 && source.indexOf('_') < 0) {
            return source;
        }
        return source.replace(/[-_][^-_]/g, function (match) {
            return match.charAt(1).toUpperCase();
        });
    };

    baidu.dom.setStyle = function (element, key, value) {
        var dom = baidu.dom, fixer;
        element = dom.g(element);
        key = baidu.string.toCamelCase(key);

        if (fixer = dom._styleFilter) {
            value = fixer.filter(key, value, 'set');
        }

        fixer = dom._styleFixer[key];
        (fixer && fixer.set) ? fixer.set(element, value) : (element.style[fixer || key] = value);

        return element;
    };

    baidu.setStyle = baidu.dom.setStyle;

    baidu.dom.setStyles = function (element, styles) {
        element = baidu.dom.g(element);
        for (var key in styles) {
            baidu.dom.setStyle(element, key, styles[key]);
        }
        return element;
    };
    baidu.setStyles = baidu.dom.setStyles;
    baidu.browser = baidu.browser || {};
    baidu.browser.ie = baidu.ie = /msie (\d+\.\d+)/i.test(navigator.userAgent) ? (document.documentMode || + RegExp['\x241']) : undefined;
    baidu.dom._NAME_ATTRS = (function () {
        var result = {
            'cellpadding': 'cellPadding',
            'cellspacing': 'cellSpacing',
            'colspan': 'colSpan',
            'rowspan': 'rowSpan',
            'valign': 'vAlign',
            'usemap': 'useMap',
            'frameborder': 'frameBorder'
        };

        if (baidu.browser.ie < 8) {
            result['for'] = 'htmlFor';
            result['class'] = 'className';
        } else {
            result['htmlFor'] = 'for';
            result['className'] = 'class';
        }

        return result;
    })();

    baidu.dom.setAttr = function (element, key, value) {
        element = baidu.dom.g(element);
        if ('style' == key) {
            element.style.cssText = value;
        } else {
            key = baidu.dom._NAME_ATTRS[key] || key;
            element.setAttribute(key, value);
        }
        return element;
    };
    baidu.setAttr = baidu.dom.setAttr;
    baidu.dom.setAttrs = function (element, attributes) {
        element = baidu.dom.g(element);
        for (var key in attributes) {
            baidu.dom.setAttr(element, key, attributes[key]);
        }
        return element;
    };
    baidu.setAttrs = baidu.dom.setAttrs;
    baidu.dom.create = function (tagName, opt_attributes) {
        var el = document.createElement(tagName),
            attributes = opt_attributes || {};
        return baidu.dom.setAttrs(el, attributes);
    };
    T.undope = true;

    var DrawingManager = function (map, opts) {

        try {
            BMap;
        } catch (e) {
            throw Error('Baidu Map JS API is not ready yet!');
        }

        if (!DrawingManager._isExtended) {
            DrawingManager._isExtended = true;

            extend(DrawingManager, BMap.Overlay, "DrawingManager");
            var drawingManager = new DrawingManager(map, opts);
            this.__proto__ = drawingManager.__proto__;
        }

        let me = this;
        me.map = map;
        me._opts = opts;
        me._drawingType = null;
        me._enableDraw = false;
        me.markerOptions = opts.markerOptions || {};
        me.circleOptions = opts.circleOptions || {};
        me.areaOptions = opts.areaOptions || {};
        me.radius = opts.circleOptions.radius;
        me.vertex = [];
        me.vertexCount = 0;
        me.polygonArea = new Map();
        me.isMoreArea = opts.areaOptions.isMoreArea || false;
        me.squareArea = opts.areaOptions.areaMaxSquare || 1400;

        EventWrapper.addListener(me.map, 'zoomend', () => {
            me._dispatchEvent(me, "draw:zoom_map", 15);
        });
    }

    baidu.object.extend(DrawingManager.prototype, {

        setDrawingMode: function (drawingType, opts) {
            var me = this;
            me.markerOptions = opts.markerOptions || {};
            me.circleOptions = opts.circleOptions || {};
            me.areaOptions = opts.areaOptions || {};
            me._enableDraw = opts.enableDraw;

            if (me.markerOptions) {
                me.setPosition(me.markerOptions.markerLatitude, me.markerOptions.markerLongitude);
            }

            if (me.circleOptions) {
                me.radius = me.circleOptions.radius;
            }

            if (me.areaOptions) {
                me.areaCoordinates = me.areaOptions.areaCoordinates;
            }

            if (me._drawingType != drawingType) {
                switch (drawingType) {
                    case DRAWING_MODE_MARKER:
                        me._bindMarker(false);
                        break;
                    case DRAWING_MODE_CIRCLE:
                        me._bindMarker(true);
                        break;
                    case DRAWING_MODE_AREA:
                        me._bindArea();
                        break;
                    case DRAWING_MODE_FITBOUNDS:
                        me._fitBounds();
                        break;
                    default:
                        me._reset();
                        break;
                }
                if (drawingType == DRAWING_MODE_FITBOUNDS || drawingType == DRAWING_MODE_AREA || drawingType == DRAWING_MODE_CIRCLE) { me._drawingType = null; return false }
                me._drawingType = drawingType;
            }
        },

        _bindMarker: function (enabledCircle) {
            let me = this;
            me._removeCenterMarker();
            me._removeCircle();
            EventWrapper.clearListeners(me.map, 'click');

            var createCenterMarker = (e) => {
                let isIconSelected = false;
                me._removeArea();
                me._removeCenterMarker();
                me._removeCircle();

                if (e) {

                    me._setPosition(e);
                }
                if (me.position) {

                    const icon = me.markerOptions.iconUrl ?
                        new BMap.Icon(me.markerOptions.iconUrl, new BMap.Size(24, 30)) : null;

                    const options = {
                        draggingCursor: 'move'
                    }

                    me._centerMarker = new BMap.Marker(me.position, options);

                    me.map.addOverlay(me._centerMarker);

                    me._centerMarker.setIcon(icon);
                    me._centerMarker.setZIndex(99999);

                    if (me.markerOptions.fitBounds) {
                        me.map.setCenter(me.position);
                        if (me.map.getZoom() < 9) {
                            me.map.setZoom(9);
                        }
                    }

                    me.position = null;
                    me._dispatchEvent(me, "draw:marker_create", me._getPosition());
                    me._centerMarker.addEventListener('click', (event) => {
                        if (!me._circle) {
                            me._dispatchEvent(me, "draw:marker_click", me._getPosition());
                            if (!isIconSelected) {
                                me._setIconMarker(me.markerOptions.iconUrlSelected);
                                isIconSelected = true;
                            }
                        }

                        if (event && event.stopPropagation) {
                            event.stopPropagation();
                        } else {
                            window.event.cancelBubble = true;
                        }
                    });

                    if (enabledCircle) {
                        me._bindCircle();
                    }
                }
            }

            if (!this._enableDraw) {
                createCenterMarker();
                EventWrapper.clearListeners(me.map, 'click');
            }

            EventWrapper.addListener(me.map, 'click', (event) => {
                if (this._enableDraw) {
                    createCenterMarker(event)
                }
            });
        },

        _fitBoundsArea: function () {
            let me = this;
            if (me._area) {
                me.map.setViewport(me._area.getBounds());
            }
        },

        _fitBoundsMarker: function () {
            let me = this;
            if (me._centerMarker) {
                me.map.setCenter(me._centerMarker.getPosition());
                if (me.map.getZoom() < 9) {
                    me.map.setZoom(9);
                }
            }
        },

        _fitBoundsCircle: function () {
            let me = this;
            if (me._circle) {
                me.map.setViewport(me._circle.getBounds());
            }
        },

        _bindCircle: function () {

            let me = this;
            if (me._circle) {
                me.map.removeOverlay(me._circle);
                me.map.removeOverlay(me._vertexMarker);
                me._dispatchEvent(me, "draw:circle_remove", null);
            }

            const options = {
                fillOpacity: me.circleOptions.fillOpacity,
                strokeOpacity: me.circleOptions.strokeOpacity,
                strokeColor: me.circleOptions.strokeColor,
                strokeWeight: me.circleOptions.strokeWeight,
                fillColor: me.circleOptions.fillColor,
            }

            if (me._centerMarker) {
                me._circle = new BMap.Circle(me._centerMarker.getPosition(), Math.floor(me.radius), options);
                me.map.addOverlay(me._circle);
                me._createVertexMarker();
                me._createVertexMarkerCopy();
                me._centerMarker.enableDragging();
                me._setIconMarker(me.markerOptions.iconUrl);
                me._centerMarkerAddEventListener();

                if (me.circleOptions.fitBounds) {
                    me._fitBoundsCircle();
                }

                me._dispatchEvent(me, "draw:circle_create", me._getInfo());

            }

        },

        _bindArea: function () {

            let me = this;
            EventWrapper.clearListeners(me.map, 'click');
            if (!me.isMoreArea) {
                me._removeArea();
            }
            me._removeCenterMarker();
            me._removeCircle();

            var createArea = () => {

                var patch = [];
                me._setDrawing(false);

                const options = {
                    strokeColor: me.areaOptions.strokeColor,
                    strokeOpacity: me.areaOptions.strokeOpacity,
                    fillColor: me.areaOptions.fillColor,
                    fillOpacity: me.areaOptions.fillOpacity,
                    strokeWeight: me.areaOptions.strokeWeight,
                }

                me._area = new BMap.Polyline([], options);
                me.map.addOverlay(me._area);

                var move = EventWrapper.addListener(me.map, 'mousemove', (event) => {
                    if (typeof event.point !== "undefined") {
                        patch.push(event.point);
                        me.map.setDefaultCursor('not-allowed');
                        let area = me.geodesicArea(patch);
                        if (area < me.squareArea * 1000000) {
                            me._area.setPath(patch);
                            me.map.setDefaultCursor('crosshair');
                            return false;
                        }
                    }
                });

                EventWrapper.addListenerOnce(me.map, 'mouseup', () => {

                    EventWrapper.removeListener(move);
                    var _patch = me._area.getPath();
                    me.map.removeOverlay(me._area);
                    if (!_patch.length) { return false };
                    me._area = new BMap.Polygon(_patch, options);
                    me.map.addOverlay(me._area);
                    me.vertexCount++;
                    EventWrapper.removeListener(mousedown);
                    me.polygonArea.set(me.vertexCount, me._area);
                    me._dispatchEvent(me, "draw:area_create", me._convertCoordinates(me._area.getPath()));
                    me._setDrawing(true);

                    if (me.areaOptions.fitBounds) {
                        me._fitBoundsArea();
                    }
                    if (me.isMoreArea) {
                        patchAreaLatLng = me._area.getPath();
                        let patchAreaLatLngMax = me.findRoundPoints(patchAreaLatLng, 0.001);
                        if (patchAreaLatLngMax) {
                            if (patchAreaLatLngMax.length == 0) {
                                patchAreaLatLngMax = me.findRoundPoints(patchAreaLatLng, 0.01);
                            }
                            if (patchAreaLatLngMax.length == 0) {
                                patchAreaLatLngMax = me.findRoundPoints(patchAreaLatLng, 1);
                            }
                            if (patchAreaLatLngMax.length == 0) {
                                patchAreaLatLngMax = patchAreaLatLng;
                            }
                        }

                        var pointSelect = null;
                        var pointSelect = patchAreaLatLngMax[0];

                        for (var n = 0; n < patchAreaLatLngMax.length; n++) {
                            let item = patchAreaLatLngMax[n];
                            if (item.lng > pointSelect.lng) {
                                pointSelect = item;
                            }
                        }

                        me.createVertexArea(pointSelect);
                    }

                });
            }


            if (this._enableDraw) {
                var mousedown = EventWrapper.addListener(me.map, 'mousedown', (e) => {
                    e = window.event || e;
                    if (e.button == 2) { return; }
                    createArea();
                });
            }
            else {
                me._redrawArea()
            }
        },

        geodesicArea: function (latLngs) {
            var pointsCount = latLngs.length,
                area = 0.0,
                d2r = Math.PI / 180,
                p1, p2;

            if (pointsCount > 2) {
                for (var i = 0; i < pointsCount; i++) {
                    p1 = latLngs[i];
                    p2 = latLngs[(i + 1) % pointsCount];
                    area += ((p2.lng - p1.lng) * d2r) *
                        (2 + Math.sin(p1.lat * d2r) + Math.sin(p2.lat * d2r));
                }
                area = area * 6378137.0 * 6378137.0 / 2.0;
            }

            return Math.abs(area);
        },

        createVertexArea: function (pointSelect) {
            let me = this;
            let icon = new BMap.Icon(me.areaOptions.areaIconVertex, new BMap.Size(24, 24), {});

            _vertex = new BMap.Marker(pointSelect, {});
            me.map.addOverlay(_vertex);
            _vertex.setIcon(icon);
            _vertex.setZIndex(99);
            _vertex.draw();
            _vertex["id"] = me.vertexCount;

            _vertex.addEventListener('click', function () {
                if (me.polygonArea.get(this.id)) {
                    me.map.removeOverlay(me.polygonArea.get(this.id));
                    me.polygonArea.delete(this.id);
                    me.map.removeOverlay(this);
                    me._dispatchEvent(me, "draw:area_remove", me._convertCoordinates());
                }
            });

            me.vertex.push(_vertex);

        },

        findRoundPoints: function (patch, round) {
            let me = this;
            var bounds = me.map.getBounds();
            patch.forEach(function (element, index) { bounds.extend(element) });
            var center = bounds.getCenter();
            var result = [];
            var roundMinLat = center.lat - round;
            var roundMaxLat = center.lat + round;
            for (var n = 0; n < patch.length; n++) {
                let item = patch[n];
                if (item.lat >= roundMinLat && item.lat <= roundMaxLat) {
                    if (item.lng > center.lng) {
                        result.push(item);
                    }
                }
            }

            return result;
        },
        _redrawArea: function () {
            let me = this;
            let latlngs = me._convertlatlngs();

            const options = {
                strokeColor: me.areaOptions.strokeColor,
                strokeOpacity: me.areaOptions.strokeOpacity,
                fillColor: me.areaOptions.fillColor,
                fillOpacity: me.areaOptions.fillOpacity,
                strokeWeight: me.areaOptions.strokeWeight,
            }

            me._area = new BMap.Polygon(latlngs, options);
            me.map.addOverlay(me._area);
            me._dispatchEvent(me, "draw:area_create", me._convertCoordinates(me._area.getPath()));
            me._fitBoundsArea();

        },

        _setIconMarker: function (iconUrl) {
            let me = this;
            if (me._centerMarker) {
                const icon = new BMap.Icon(iconUrl, new BMap.Size(24, 30));
                me._centerMarker.setIcon(icon);
            }
        },

        _convertlatlngs: function () {
            let me = this;
            let latlngs = [];

            if (me.areaCoordinates.length > 0) {
                for (var n = 0; n < me.areaCoordinates.length; n++) {
                    let item = me.areaCoordinates[n];
                    let position = new BMap.Point(item.longitude, item.latitude);
                    latlngs.push(position);
                }
                return latlngs;
            }
        },

        _reset: function () {
            let me = this;
            me._removeArea();
            me._removeCenterMarker();
            me._removeCircle();
        },

        _fitBounds: function () {
            let me = this;
            if (me._centerMarker) {
                me._fitBoundsMarker();
            }
            if (me._area) {
                me._fitBoundsArea();

            }
            if (me._circle) {
                me._fitBoundsCircle();
            }
        },

        _setDrawing: function (enabled) {
            let me = this;
            enabled ? me.map.enableDragging() : me.map.disableDragging();
            enabled ? me.map.enableScrollWheelZoom() : me.map.disableScrollWheelZoom();
            enabled ? me.map.enableDoubleClickZoom() : me.map.disableDoubleClickZoom();

        },

        _convertCoordinates: function () {
            let me = this;
            let positions = [];
            let arrayArea = {};
            let id = null;
            if (!me.isMoreArea) {
                for (var n = 0; n < me._area.getPath().length; n++) {
                    let item = me._area.getPath()[n];
                    let position = {
                        latitude: item.lat,
                        longitude: item.lng,
                    }
                    positions.push(position);
                    id = 0;
                }
                arrayArea = {
                    id: id,
                    positions,
                }
                return { arrayArea };

            }
            else {
                me.polygonArea.forEach((key, value) => {
                    let latLng = [];
                    for (var n = 0; n < key.getPath().length; n++) {
                        let item = key.getPath()[n];
                        let position = {
                            latitude: item.lat,
                            longitude: item.lng,
                        }
                        latLng.push(position);
                        id = value;
                    }
                    positions.push(latLng);

                });
            }

            arrayArea = {
                id: id,
                positions,
            }
            return { arrayArea };

        },

        _getPosition: function () {
            let me = this;
            let position = {
                latitude: me._centerMarker.getPosition().lat,
                longitude: me._centerMarker.getPosition().lng
            }

            let pos = { position };
            return pos;
        },

        setPosition: function (latitude, longitude) {
            let me = this;
            me.position = null;
            me.position = new BMap.Point(longitude, latitude);
        },

        _setPosition: function (e) {
            let me = this;
            me.position = null;
            me.position = e.point;
        },

        setEnableDraw: function (enabled) {
            this._enableDraw = enabled;
        },

        _removeCircle: function () {
            let me = this;
            if (me._circle) {
                me.map.removeOverlay(me._circle);
                me.map.removeOverlay(me._vertexMarker);
                me._circle = null;
                me._vertexMarker = null;
                me._dispatchEvent(me, "draw:circle_remove", null);

                if (me._vertexMarkerCopy) {
                    me.map.removeOverlay(me._vertexMarkerCopy);
                    me._vertexMarkerCopy = null;
                }
            }
        },

        _removeCenterMarker: function () {
            let me = this;
            if (me._centerMarker) {
                me.map.removeOverlay(me._centerMarker);
                me._centerMarker = null;
                me._dispatchEvent(me, "draw:marker_remove", null);
            }

        },

        _removeArea: function () {
            let me = this;
            if (me._area) {
                me.map.removeOverlay(me._area);
                me._area = null;
            }

            me.polygonArea.forEach((key) => {
                me.map.removeOverlay(key);
            });

            me.vertex.forEach((key) => {
                me.map.removeOverlay(key);
            });

            me.polygonArea.clear();
            me.vertex = [];
            me.currentAreaCount = 0;
            me._dispatchEvent(me, "draw:area_remove", []);

        },

        _getInfo: function () {
            let me = this;
            let position = {
                latitude: me._centerMarker.getPosition().lat,
                longitude: me._centerMarker.getPosition().lng
            }
            let info = {
                radius: Math.floor(me._circle.getRadius()),
                position,
            };

            return info;
        },

        _getZoom: function () {
            let me = this;
            let zoom = {
                zoom: me.map.getZoom()
            }
            return zoom;
        },

        _vertexMarkerAddEventListener: function () {
            let me = this;
            me.isMaxDistance = false;
            me._vertexMarker.addEventListener('dragging', (event) => {
                me.isMaxDistance = false;
                let distance = me.getDistanceTo(me._centerMarker.getPosition(), event.point);
                if (distance >= me.circleOptions.maxRadius) {

                    me._vertexMarker.hide();
                    me._vertexMarker.setZIndex(100);

                    if (me._vertexMarkerCopy) {
                        me._vertexMarkerCopy.show();
                        me._vertexMarkerCopy.setZIndex(200);
                    }

                    me.isMaxDistance = true;
                    me._dispatchEvent(me, "draw:circle_radius_max", null);
                    me.map.setDefaultCursor('not-allowed');

                    return false;
                }


                me.radius = distance;
                if (me._circle) {
                    me._circle.setRadius(distance);
                }

                if (me._vertexMarkerCopy) {
                    me.map.removeOverlay(me._vertexMarkerCopy);
                }

                if (!me.isMaxDistance) {
                    me._createVertexMarkerCopy();
                }

                me._vertexMarker.show();
                me._vertexMarker.setZIndex(200);

                if (me._vertexMarkerCopy) {
                    me._vertexMarkerCopy.hide();
                    me._vertexMarkerCopy.setZIndex(100);
                }

                let pixel = {
                    clientX: event.pixel.x,
                    clientY: event.pixel.y,
                }
                let ev = {
                    pixel,
                    radius: Math.floor(me._circle.getRadius()),
                }

                me._dispatchEvent(me, "draw:circle_radius_change", ev);


                me._vertexMarker.addEventListener('mouseover', () => {

                    if (me._vertexMarkerCopy) {
                        me._vertexMarkerCopy.hide();
                        me._vertexMarkerCopy.setZIndex(100);
                    }

                    me._vertexMarker.show();
                    me._vertexMarker.setZIndex(200);
                    me._vertexMarker.removeEventListener('mouseover');
                });

            });

            me._vertexMarker.addEventListener('dragend', function () {
                me.map.setDefaultCursor('inherit');
                if (me._vertexMarkerCopy) {
                    if (me.isMaxDistance) {

                        me._vertexMarker.setPosition(me._vertexMarkerCopy.getPosition());
                        if (me._vertexMarkerCopy) {
                            me._vertexMarkerCopy.show();
                            me._vertexMarkerCopy.setZIndex(200);
                        }

                        me._vertexMarker.hide();
                        me._vertexMarker.setZIndex(100);
                    }
                }

                let position = {
                    latitude: me._centerMarker.getPosition().lat,
                    longitude: me._centerMarker.getPosition().lng
                }
                let info = {
                    radius: me._circle.getRadius(),
                    position,
                    latLng: me._centerMarker.getPosition()
                };

                me._dispatchEvent(me, "draw:circle_radius_complete", info);
            });
        },

        _vertexMarkerCopyAddEventListener: function () {
            let me = this;
            me._vertexMarkerCopy.addEventListener('mouseover', () => {

                if (me._vertexMarkerCopy) {
                    me._vertexMarkerCopy.hide();
                    me._vertexMarkerCopy.setZIndex(100);
                }

                me._vertexMarker.show();
                me._vertexMarker.setZIndex(200);

                me._vertexMarkerCopy.removeEventListener('mouseover')
            });
        },

        _centerMarkerAddEventListener: function () {
            let me = this;
            me._centerMarker.addEventListener('dragging', function (event) {
                if (me._vertexMarkerCopy) {
                    me._vertexMarkerCopy.hide();
                    me._vertexMarkerCopy.setZIndex(100);
                    me._vertexMarker.hide();
                    me._vertexMarker.setZIndex(200);
                }

                me._circle.setCenter(event.point);

                let to = me.destination(event.point, 90, me.radius);

                if (me._vertexMarker) {
                    me._vertexMarker.setPosition(to);
                }
                if (me._vertexMarkerCopy) {
                    me._vertexMarkerCopy.setPosition(to);
                };

                me._dispatchEvent(me, "draw:circle_centre_change", me._getInfo());
            });

            me._centerMarker.addEventListener('dragend', () => {
                me._vertexMarker.show();
                me._dispatchEvent(me, "draw:circle_center_complete", me._getInfo());

            });

            me._centerMarker.addEventListener('mouseout', () => {
                me._dispatchEvent(me, "draw:marker_mouseout", me._getInfo());
            });

            me._centerMarker.addEventListener('mouseover', () => {
                me._dispatchEvent(me, "draw:marker_mouseover", me._getInfo());

            });
            me._centerMarker.removeEventListener('click');
        },

        _createVertexMarker: function () {

            let me = this;
            this.to = null;
            this.to = me.destination(this._centerMarker.getPosition(), 90, this.radius);

            if (me._vertexMarker) {
                me.map.removeOverlay(me._vertexMarker);
            }

            const options = {
                draggingCursor: 'col-resize'
            }

            me._vertexMarker = new BMap.Marker(me.to, options);

            me.map.addOverlay(me._vertexMarker);
            me._vertexMarker.enableDragging();

            const svg = [
                '<?xml version="1.0"?>',
                '<svg width="15px" height="15px" viewBox="0 0 100 100" version="1.1" xmlns="http://www.w3.org/2000/svg">',
                '<circle stroke="#003dd9" fill="white" stroke-width="10" cx="50" cy="50" r="35"/>',
                '</svg>'
            ].join('\n');

            const iconU = 'data:image/svg+xml;charset=UTF-8,' + encodeURIComponent(svg);

            const icon = me.markerOptions.iconUrl ?
                new BMap.Icon(iconU, new BMap.Size(15, 15)) : null;
            me._vertexMarker.setIcon(icon);
            me._vertexMarkerAddEventListener();

        },

        _createVertexMarkerCopy: function () {
            let me = this;
            me._vertexMarkerCopy = new BMap.Marker(me._vertexMarker.getPosition(), {});
            me.map.addOverlay(me._vertexMarkerCopy);

            const svg = [
                '<?xml version="1.0"?>',
                '<svg width="15px" height="15px" viewBox="0 0 100 100" version="1.1" xmlns="http://www.w3.org/2000/svg">',
                '<circle stroke="#003dd9" fill="white" stroke-width="10" cx="50" cy="50" r="35"/>',
                '</svg>'
            ].join('\n');

            const iconU = 'data:image/svg+xml;charset=UTF-8,' + encodeURIComponent(svg);
            const icon = me.markerOptions.iconUrl ?
                new BMap.Icon(iconU, new BMap.Size(15, 15)) : null;
            me._vertexMarkerCopy.setIcon(icon);
            me._vertexMarkerCopyAddEventListener();
        },

        destination: function (latlng, heading, distance) {

            heading = (heading + 360) % 360;
            var rad = Math.PI / 180,
                radInv = 180 / Math.PI,
                R = 6378137,
                lon1 = latlng.lng * rad,
                lat1 = latlng.lat * rad,
                rheading = heading * rad,
                sinLat1 = Math.sin(lat1),
                cosLat1 = Math.cos(lat1),
                cosDistR = Math.cos(distance / R),
                sinDistR = Math.sin(distance / R),
                lat2 = Math.asin(sinLat1 * cosDistR + cosLat1 *
                    sinDistR * Math.cos(rheading)),
                lon2 = lon1 + Math.atan2(Math.sin(rheading) * sinDistR *
                    cosLat1, cosDistR - sinLat1 * Math.sin(lat2));
            lon2 = lon2 * radInv;
            lon2 = lon2 > 180 ? lon2 - 360 : lon2 < -180 ? lon2 + 360 : lon2;
            return new BMap.Point(lon2, lat2 * radInv);
        }
        ,

        degreeToRad: function (degree) {
            return Math.PI * degree / 180;
        },

        _getRange: function (v, a, b) {
            if (a != null) {
                v = Math.max(v, a);
            }
            if (b != null) {
                v = Math.min(v, b);
            }
            return v;
        },

        _getLoop: function (v, a, b) {
            while (v > b) {
                v -= b - a
            }
            while (v < a) {
                v += b - a
            }
            return v;
        },

        getDistanceTo: function (point1, point2) {
            let me = this;
            if (!(point1 instanceof BMap.Point) ||
                !(point2 instanceof BMap.Point)) {
                return 0;
            }

            point1.lng = me._getLoop(point1.lng, -180, 180);
            point1.lat = me._getRange(point1.lat, -74, 74);
            point2.lng = me._getLoop(point2.lng, -180, 180);
            point2.lat = me._getRange(point2.lat, -74, 74);

            var x1, x2, y1, y2;
            x1 = this.degreeToRad(point1.lng);
            y1 = this.degreeToRad(point1.lat);
            x2 = this.degreeToRad(point2.lng);
            y2 = this.degreeToRad(point2.lat);

            return 6370996.81 * Math.acos((Math.sin(y1) * Math.sin(y2) + Math.cos(y1) * Math.cos(y2) * Math.cos(x2 - x1)));
        },

        _stopBubble: function (e) {
            if (e && e.stopPropagation) {
                e.stopPropagation();
            } else {
                window.event.cancelBubble = true;
            }
        },

        _dispatchEvent: function (instance, type, opts) {
            type.indexOf("on") != 0 && (type = "on" + type);
            var event = new baidu.lang.Event(type);
            if (!!opts) {
                for (var p in opts) {
                    event[p] = opts[p];
                }
            }
            instance.dispatchEvent(event);
        }
    });
    return DrawingManager;
});